/* Author : Dario GENGA
 * Date : 16.11.2021
 * Description : Manipulate matrix
 */

#include "matrix.h"
#include <stdio.h>

error_code matrix_alloc(matrix *mat, int32_t m, int32_t n) {
    if (m <= 0 || n <= 0) {
        return err;
    }

    int32_t **data = malloc(m * sizeof(int32_t *));
    mat->m = m;
    mat->n = n;

    if (data == NULL) {
        return err;
    }
    mat->data = data;

    for (int32_t i = 0; i < m; i++) {
        int32_t *row = malloc(n * sizeof(int32_t *));

        if (row == NULL) {
            return err;
        }

        mat->data[i] = row;
    }

    return ok;
}

error_code matrix_init(matrix *mat, int32_t m, int32_t n, int32_t val) {
    if (m <= 0 || n <= 0) {
        return err;
    }

    matrix_alloc(mat, m, n);

    for (int32_t i = 0; i < m; i++) {
        for (int32_t k = 0; k < n; k++) {
            mat->data[i][k] = val;
        }
    }

    return ok;
}

error_code matrix_init_from_array(matrix *mat, int32_t m, int32_t n, int32_t data[], int32_t s) {
    if (m <= 0 || n <= 0 || m * n != s) {
        return err;
    }

    matrix_alloc(mat, m, n);

    for (int i = 0; i < m; i++) {
        for (int k = 0; k < n; k++) {
            mat->data[i][k] = data[k + (i * n)];
        }
    }

    return ok;
}

error_code matrix_print(const matrix mat) {
    for (int i = 0; i < mat.m; i++) {
        printf("[");
        for (int k = 0; k < mat.n; k++) {
            printf(" %d ", mat.data[i][k]);
        }
        printf("]\n");
    }

    return ok;
}

error_code matrix_clone(matrix *cloned, const matrix mat) {
    error_code code = matrix_alloc(cloned, mat.m, mat.n);
    if (code == err) {
        return err;
    }

    for (int i = 0; i < mat.m; i++) {
        for (int k = 0; k < mat.n; k++) {
            cloned->data[i][k] = mat.data[i][k];
        }
    }

    return ok;
}

error_code matrix_destroy(matrix *mat) {
    // Free each row
    for (int32_t i = 0; i < mat->m; i++) {
        free(mat->data[i]);
    }

    // Free the data and reset the number of row and col
    free(mat->data);
    mat->m = -1;
    mat->n = -1;
    mat->data = NULL;

    return ok;
}

error_code matrix_transpose(matrix *transposed, const matrix mat) {
    error_code code = matrix_alloc(transposed, mat.n, mat.m);
    if (code == err) {
        return err;
    }

    for (int i = 0; i < mat.m; i++) {
        for (int k = 0; k < mat.n; k++) {
            transposed->data[k][i] = mat.data[i][k];
        }
    }

    return ok;
}

error_code matrix_extract_submatrix(matrix *sub, const matrix mat, int32_t m0, int32_t m1, int32_t n0, int32_t n1) {
    if (m0 < 0 || m1 > mat.m || n0 < 0 || n1 > mat.n || m0 > m1 || n0 > n1) {
        return err;
    }

    error_code code = matrix_alloc(sub, m1 - m0 , n1 - n0 );
    if (code == err) {
        return err;
    }

    for (int i = m0; i < m1; i++) {
        for (int k = n0; k < n1; k++) {
            sub->data[i - m0][k - n0] = mat.data[i][k];
        }
    }

    return ok;
}


bool matrix_is_equal(matrix mat1, matrix mat2) {
    if (mat1.m != mat2.m || mat1.n != mat2.n) {
        return false;
    }

    // Parse each values of the matrix
    for (int i = 0; i < mat1.m; i++) {
        for (int k = 0; k < mat1.n; k++) {
            if (mat1.data[i][k] != mat2.data[i][k]) {
                return false;
            }
        }
    }
    return true;
}

error_code matrix_get(int32_t *elem, const matrix mat, int32_t ix, int32_t iy) {
    if (ix < 0 || iy < 0 || iy >= mat.m || ix >= mat.n) {
        return err;
    }
    *elem = mat.data[iy][ix];

    return ok;
}


error_code matrix_set(matrix mat, int32_t ix, int32_t iy, int32_t elem) {
    if (ix < 0 || iy < 0 || ix > mat.m || iy > mat.n) {
        return err;
    }
    mat.data[iy][ix] = elem;

    return ok;
}
